@@ -1,61 +1,22 @@
 /*
+ * For HPND-sell-variant:
+ * Copyright 1992 by Oki Technosystems Laboratory, Inc.
+ * Copyright 1992 by Fuji Xerox Co., Ltd.
+ *
+ * For MIT:
  * Copyright © 2013 Ran Benita <ran234@gmail.com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
+ * SPDX-License-Identifier: HPND-sell-variant AND MIT
  *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * Author: Yasuhiro Kawai        Oki Technosystems Laboratory
+ * Author: Kazunori Nishihara    Fuji Xerox
  */
 
-/******************************************************************
-
-              Copyright 1992 by Oki Technosystems Laboratory, Inc.
-              Copyright 1992 by Fuji Xerox Co., Ltd.
-
-Permission to use, copy, modify, distribute, and sell this software
-and its documentation for any purpose is hereby granted without fee,
-provided that the above copyright notice appear in all copies and
-that both that copyright notice and this permission notice appear
-in supporting documentation, and that the name of Oki Technosystems
-Laboratory and Fuji Xerox not be used in advertising or publicity
-pertaining to distribution of the software without specific, written
-prior permission.
-Oki Technosystems Laboratory and Fuji Xerox make no representations
-about the suitability of this software for any purpose.  It is provided
-"as is" without express or implied warranty.
-
-OKI TECHNOSYSTEMS LABORATORY AND FUJI XEROX DISCLAIM ALL WARRANTIES
-WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
-MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL OKI TECHNOSYSTEMS
-LABORATORY AND FUJI XEROX BE LIABLE FOR ANY SPECIAL, INDIRECT OR
-CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
-OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
-OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
-OR PERFORMANCE OF THIS SOFTWARE.
-
-  Author: Yasuhiro Kawai        Oki Technosystems Laboratory
-  Author: Kazunori Nishihara    Fuji Xerox
-
-******************************************************************/
-
 #include "config.h"
 
 #include <errno.h>
 
+#include "messages-codes.h"
 #include "utils.h"
 #include "table.h"
 #include "scanner-utils.h"
@@ -146,12 +107,12 @@ skip_more_whitespace_and_comments:
         while (scanner_peek(s) != '>' && !scanner_eol(s) && !scanner_eof(s))
             scanner_buf_append(s, scanner_next(s));
         if (!scanner_chr(s, '>')) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "unterminated keysym literal");
             return TOK_ERROR;
         }
         if (!scanner_buf_append(s, '\0')) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "keysym literal is too long");
             return TOK_ERROR;
         }
@@ -185,8 +146,8 @@ skip_more_whitespace_and_comments:
                         scanner_buf_append(s, (char) o);
                     } else {
                         scanner_warn(s, XKB_WARNING_INVALID_ESCAPE_SEQUENCE,
-                                     "illegal hexadecimal escape sequence (%.*s) "
-                                     "in string literal",
+                                     "illegal hexadecimal escape sequence "
+                                     "\"%.*s\" in string literal",
                                      (int) (s->pos - start_pos + 1),
                                      &s->s[start_pos - 1]);
                     }
@@ -196,7 +157,7 @@ skip_more_whitespace_and_comments:
                 }
                 else if (s->pos > start_pos) {
                     scanner_warn(s, XKB_WARNING_INVALID_ESCAPE_SEQUENCE,
-                                 "illegal octal escape sequence (%.*s) "
+                                 "illegal octal escape sequence \"%.*s\" "
                                  "in string literal",
                                  (int) (s->pos - start_pos + 1),
                                  &s->s[start_pos - 1]);
@@ -204,7 +165,8 @@ skip_more_whitespace_and_comments:
                 }
                 else {
                     scanner_warn(s, XKB_WARNING_UNKNOWN_CHAR_ESCAPE_SEQUENCE,
-                                 "unknown escape sequence (\\%c) in string literal",
+                                 "unknown escape sequence \"\\%c\" "
+                                 "in string literal",
                                  scanner_peek(s));
                     /* Ignore. */
                 }
@@ -213,17 +175,17 @@ skip_more_whitespace_and_comments:
             }
         }
         if (!scanner_chr(s, '\"')) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "unterminated string literal");
             return TOK_ERROR;
         }
         if (!scanner_buf_append(s, '\0')) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "string literal is too long");
             return TOK_ERROR;
         }
         if (!is_valid_utf8(s->buf, s->buf_pos - 1)) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_FILE_ENCODING,
                         "string literal is not a valid UTF-8 string");
             return TOK_ERROR;
         }
@@ -238,7 +200,7 @@ skip_more_whitespace_and_comments:
         while (is_alnum(scanner_peek(s)) || scanner_peek(s) == '_')
             scanner_buf_append(s, scanner_next(s));
         if (!scanner_buf_append(s, '\0')) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "identifier is too long");
             return TOK_ERROR;
         }
@@ -251,7 +213,7 @@ skip_more_whitespace_and_comments:
         return TOK_IDENT;
     }
 
-    scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+    scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                 "unrecognized token");
     /* Discard rest of line. */
     scanner_skip_to_eol(s);
@@ -270,7 +232,7 @@ lex_include_string(struct scanner *s, struct xkb_compose_table *table,
     s->buf_pos = 0;
 
     if (!scanner_chr(s, '\"')) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                     "include statement must be followed by a path");
         return TOK_ERROR;
     }
@@ -318,7 +280,7 @@ lex_include_string(struct scanner *s, struct xkb_compose_table *table,
                 }
             }
             else {
-                scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+                scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                             "unknown %% format (%c) in include statement", scanner_peek(s));
                 return TOK_ERROR;
             }
@@ -327,7 +289,7 @@ lex_include_string(struct scanner *s, struct xkb_compose_table *table,
         }
     }
     if (!scanner_chr(s, '\"')) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                     "unterminated include statement");
         return TOK_ERROR;
     }
@@ -502,7 +464,7 @@ parse_string_literal(struct xkb_context *ctx, const char *string)
 {
     struct scanner s;
     union lvalue val;
-    scanner_init(&s, ctx, string, strlen(string), "(unamed)", NULL);
+    scanner_init(&s, ctx, string, strlen(string), "(unnamed)", NULL);
     switch (lex(&s, &val)) {
         case TOK_STRING:
             return strdup(val.string.str);
@@ -577,7 +539,7 @@ parse(struct xkb_compose_table *table, struct scanner *s,
        The first character relevant to the grammar must be ASCII:
        whitespace, include, modifier list, keysym, comment */
     if (!scanner_check_supported_char_encoding(s)) {
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_FILE_ENCODING,
                     "This could be a file encoding issue. "
                     "Supported file encodings are ASCII and UTF-8.");
         goto fail;
@@ -665,7 +627,7 @@ lhs_keysym_tok:
         check_deprecated_keysyms(scanner_warn, s, s->ctx,
                                  keysym, val.string.str, val.string.str, "%s", "");
         if (production.len + 1 > MAX_LHS_LEN) {
-            scanner_warn(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_warn(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                          "too many keysyms (%d) on left-hand side; skipping line",
                          MAX_LHS_LEN + 1);
             goto skip;
@@ -697,7 +659,7 @@ lhs_mod_list_tok: {
 
         mod = resolve_modifier(val.string.str);
         if (mod == XKB_MOD_INVALID) {
-            scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                         "unrecognized modifier \"%s\"",
                         val.string.str);
             goto error;
@@ -749,7 +711,7 @@ rhs:
         check_deprecated_keysyms(scanner_warn, s, s->ctx,
                                  keysym, val.string.str, val.string.str, "%s", "");
         if (production.has_keysym) {
-            scanner_warn(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_warn(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                          "right-hand side can have at most one keysym; "
                          "skipping line");
             goto skip;
@@ -759,7 +721,7 @@ rhs:
         /* fallthrough */
     case TOK_END_OF_LINE:
         if (!production.has_string && !production.has_keysym) {
-            scanner_warn(s, XKB_LOG_MESSAGE_NO_ID,
+            scanner_warn(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                          "right-hand side must have at least one of string "
                          "or keysym; skipping line");
             goto skip;
@@ -772,7 +734,7 @@ rhs:
 
 unexpected:
     if (tok != TOK_ERROR)
-        scanner_err(s, XKB_LOG_MESSAGE_NO_ID,
+        scanner_err(s, XKB_ERROR_INVALID_COMPOSE_SYNTAX,
                     "unexpected token");
 error:
     num_errors++;
