@@ -853,7 +853,7 @@ bin_search(const struct codepair *table, size_t length, xkb_keysym_t keysym)
     return NO_KEYSYM_UNICODE_CONVERSION;
 }
 
-XKB_EXPORT uint32_t
+uint32_t
 xkb_keysym_to_utf32(xkb_keysym_t keysym)
 {
     /* first check for Latin-1 characters (1:1 mapping) */
@@ -878,8 +878,9 @@ xkb_keysym_to_utf32(xkb_keysym_t keysym)
     /* Exclude surrogates: they are invalid in UTF-32.
      * See https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf#G28875
      * for further details.
-    */
-    if (0x0100d800 <= keysym && keysym <= 0x0100dfff)
+     */
+    if (keysym >= XKB_KEYSYM_UNICODE_SURROGATE_MIN &&
+        keysym <= XKB_KEYSYM_UNICODE_SURROGATE_MAX)
         return NO_KEYSYM_UNICODE_CONVERSION;
     /*
      * In theory, this is supposed to start from 0x100100, such that the ASCII
@@ -894,7 +895,7 @@ xkb_keysym_to_utf32(xkb_keysym_t keysym)
     return bin_search(keysymtab, ARRAY_SIZE(keysymtab) - 1, keysym);
 }
 
-XKB_EXPORT xkb_keysym_t
+xkb_keysym_t
 xkb_utf32_to_keysym(uint32_t ucs)
 {
     /* first check for Latin-1 characters (1:1 mapping) */
@@ -909,10 +910,12 @@ xkb_utf32_to_keysym(uint32_t ucs)
     if (ucs == (XKB_KEY_Delete & 0x7f))
         return XKB_KEY_Delete;
 
-    /* Unicode non-symbols and code points outside Unicode planes */
-    if ((ucs >= 0xd800 && ucs <= 0xdfff) ||
-        (ucs >= 0xfdd0 && ucs <= 0xfdef) ||
-        ucs > 0x10ffff || (ucs & 0xfffe) == 0xfffe)
+    /* U+0000 NULL and invalid Unicode code points.
+     * Surrogates are invalid in UTF-32.
+     * See https://www.unicode.org/versions/Unicode15.0.0/ch03.pdf#G28875
+     * for further details.
+     */
+    if (unlikely(ucs == 0 || is_surrogate(ucs) || ucs > 0x10ffff))
         return XKB_KEY_NoSymbol;
 
     /* search main table */
@@ -920,36 +923,19 @@ xkb_utf32_to_keysym(uint32_t ucs)
         if (keysymtab[i].ucs == ucs)
             return keysymtab[i].keysym;
 
-    /* Use direct encoding if everything else fails */
+    /* Use direct encoding if everything else failed */
     return ucs | XKB_KEYSYM_UNICODE_OFFSET;
 }
 
 /*
+ * SPDX-SnippetBegin
  * Copyright © 2012 Intel Corporation
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  *
  * Author: Rob Bradford <rob@linux.intel.com>
  */
 
-XKB_EXPORT int
+int
 xkb_keysym_to_utf8(xkb_keysym_t keysym, char *buffer, size_t size)
 {
     uint32_t codepoint;
@@ -966,3 +952,5 @@ xkb_keysym_to_utf8(xkb_keysym_t keysym, char *buffer, size_t size)
 
     return utf32_to_utf8(codepoint, buffer);
 }
+
+/* SPDX-SnippetEnd */
