@@ -34,9 +34,10 @@
  * xkb_symbols "basic" { ... }) , as opposed to the default map of the file.
  *
  * @param nextop_rtrn Set to the next operation in the complete statement,
- * which is '\0' if it's the last file or '+' or '|' if there are more.
- * Separating the files with '+' sets the merge mode to MERGE_MODE_OVERRIDE,
- * while '|' sets the merge mode to MERGE_MODE_AUGMENT.
+ * which is '\0' if it's the last file or '+' or '|' or '^' if there are more.
+ * Separating the files with '+' sets the merge mode to `MERGE_MODE_OVERRIDE`,
+ * while '|' sets the merge mode to `MERGE_MODE_AUGMENT` and '^' sets the merge
+ * mode to `MERGE_REPLACE`.
  *
  * @param extra_data Set to the string after ':', if any. Currently the
  * extra data is only used for setting an explicit group index for a symbols
@@ -71,7 +72,7 @@ ParseIncludeMap(char **str_inout, char **file_rtrn, char **map_rtrn,
      * Find the position in the string where the next file is included,
      * if there is more than one left in the statement.
      */
-    next = strpbrk(str, "|+");
+    next = strpbrk(str, MERGE_MODE_PREFIXES);
     if (next) {
         /* Got more files, this function will be called again. */
         *nextop_rtrn = *next;
@@ -258,11 +259,18 @@ XkbFile *
 ProcessIncludeFile(struct xkb_context *ctx, IncludeStmt *stmt,
                    enum xkb_file_type file_type)
 {
-    FILE *file;
-    XkbFile *xkb_file = NULL;
+    /*
+     * Resolve include statement:
+     * - Explicit map name: look for an *exact match* only.
+     * - Default map: look for an *explicit default* map (i.e. tagged with
+     *   `default`: exact match), else fallback to the first *implicit* default
+     *   map (weak match).
+     */
+    XkbFile *xkb_file = NULL;  /* Exact match */
+    XkbFile *candidate = NULL; /* Weak match */
     unsigned int offset = 0;
 
-    file = FindFileInXkbPath(ctx, stmt->file, file_type, NULL, &offset);
+    FILE* file = FindFileInXkbPath(ctx, stmt->file, file_type, NULL, &offset);
     if (!file)
         return NULL;
 
@@ -279,8 +287,27 @@ ProcessIncludeFile(struct xkb_context *ctx, IncludeStmt *stmt,
                         xkb_file_type_to_string(xkb_file->file_type), stmt->file);
                 FreeXkbFile(xkb_file);
                 xkb_file = NULL;
-            } else {
+            } else if (stmt->map || (xkb_file->flags && MAP_IS_DEFAULT)) {
+                /*
+                 * Exact match: explicit map name or explicit default map.
+                 * Lookup stops here.
+                 */
                 break;
+            } else if (!candidate) {
+                /*
+                 * Weak match: looking for an explicit default map, but found
+                 * only the first *implicit* default map (e.g. first map of the
+                 * file). Use as fallback, but keep looking for an exact match.
+                 */
+                candidate = xkb_file;
+                xkb_file = NULL;
+            } else {
+                /*
+                 * Weak match, but we already have a previous candidate.
+                 * Keep looking for an exact match.
+                 */
+                FreeXkbFile(xkb_file);
+                xkb_file = NULL;
             }
         }
 
@@ -288,6 +315,14 @@ ProcessIncludeFile(struct xkb_context *ctx, IncludeStmt *stmt,
         file = FindFileInXkbPath(ctx, stmt->file, file_type, NULL, &offset);
     }
 
+    if (!xkb_file) {
+        /* No exact match: use weak match, if any */
+        xkb_file = candidate;
+    } else {
+        /* Found exact match: discard weak match, if any */
+        FreeXkbFile(candidate);
+    }
+
     if (!xkb_file) {
         if (stmt->map)
             log_err(ctx, XKB_ERROR_INVALID_INCLUDED_FILE,
