@@ -1,59 +1,26 @@
-/************************************************************
+/*
+ * For HPND:
  * Copyright (c) 1994 by Silicon Graphics Computer Systems, Inc.
  *
- * Permission to use, copy, modify, and distribute this
- * software and its documentation for any purpose and without
- * fee is hereby granted, provided that the above copyright
- * notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting
- * documentation, and that the name of Silicon Graphics not be
- * used in advertising or publicity pertaining to distribution
- * of the software without specific prior written permission.
- * Silicon Graphics makes no representation about the suitability
- * of this software for any purpose. It is provided "as is"
- * without any express or implied warranty.
- *
- * SILICON GRAPHICS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
- * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
- * AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SILICON
- * GRAPHICS BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
- * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
- * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
- * THE USE OR PERFORMANCE OF THIS SOFTWARE.
- *
- ********************************************************/
-
-/*
+ * For MIT:
  * Copyright © 2012 Intel Corporation
  * Copyright © 2012 Ran Benita <ran234@gmail.com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: HPND AND MIT
  *
  * Author: Daniel Stone <daniel@fooishbar.org>
- *         Ran Benita <ran234@gmail.com>
+ * Author: Ran Benita <ran234@gmail.com>
  */
 
 #include "config.h"
 
+#include "xkbcommon/xkbcommon.h"
+#include "xkbcommon/xkbcommon-keysyms.h"
+
+#include "darray.h"
+#include "keymap.h"
 #include "xkbcomp-priv.h"
+#include "darray.h"
 #include "text.h"
 #include "expr.h"
 #include "action.h"
@@ -61,6 +28,7 @@
 #include "include.h"
 #include "keysym.h"
 #include "util-mem.h"
+#include "xkbcomp/ast.h"
 
 
 enum key_repeat {
@@ -104,21 +72,10 @@ typedef struct {
     xkb_layout_index_t out_of_range_group_number;
 } KeyInfo;
 
-static void
-ClearLevelInfo(struct xkb_level *leveli)
-{
-    if (leveli->num_syms > 1) {
-        free(leveli->s.syms);
-    }
-    if (leveli->num_actions > 1) {
-        free(leveli->a.actions);
-    }
-}
-
 static void
 StealLevelInfo(struct xkb_level *into, struct xkb_level *from)
 {
-    ClearLevelInfo(into);
+    clear_level(into);
     if (from->num_syms > 1) {
         /* Steal */
         into->s.syms = steal(&from->s.syms);
@@ -148,7 +105,7 @@ ClearGroupInfo(GroupInfo *groupi)
 {
     struct xkb_level *leveli;
     darray_foreach(leveli, groupi->levels)
-        ClearLevelInfo(leveli);
+        clear_level(leveli);
     darray_free(groupi->levels);
 }
 
@@ -179,7 +136,6 @@ static void
 InitKeyInfo(struct xkb_context *ctx, KeyInfo *keyi)
 {
     memset(keyi, 0, sizeof(*keyi));
-    keyi->merge = MERGE_OVERRIDE;
     keyi->name = xkb_atom_intern_literal(ctx, "*");
     keyi->out_of_range_group_action = RANGE_WRAP;
 }
@@ -198,8 +154,8 @@ ClearKeyInfo(KeyInfo *keyi)
 typedef struct {
     enum merge_mode merge;
     bool haveSymbol;
-    // NOTE: Can also be XKB_MOD_NONE, meaning
-    //       “don’t add a modifier to the modmap”.
+    /* NOTE: Can also be XKB_MOD_NONE, meaning
+     *       “don’t add a modifier to the modmap”. */
     xkb_mod_index_t modifier;
     union {
         xkb_atom_t keyName;
@@ -211,11 +167,10 @@ typedef struct {
     char *name;         /* e.g. pc+us+inet(evdev) */
     int errorCount;
     unsigned int include_depth;
-    enum merge_mode merge;
     xkb_layout_index_t explicit_group;
     darray(KeyInfo) keys;
     KeyInfo default_key;
-    ActionsInfo *actions;
+    ActionsInfo default_actions;
     darray(xkb_atom_t) group_names;
     darray(ModMapEntry) modmaps;
     struct xkb_mod_set mods;
@@ -227,17 +182,15 @@ typedef struct {
 
 static void
 InitSymbolsInfo(SymbolsInfo *info, const struct xkb_keymap *keymap,
-                unsigned int include_depth,
-                ActionsInfo *actions, const struct xkb_mod_set *mods)
+                unsigned int include_depth, const struct xkb_mod_set *mods)
 {
     memset(info, 0, sizeof(*info));
     info->ctx = keymap->ctx;
     info->include_depth = include_depth;
     info->keymap = keymap;
-    info->merge = MERGE_OVERRIDE;
     InitKeyInfo(keymap->ctx, &info->default_key);
-    info->actions = actions;
-    info->mods = *mods;
+    InitActionsInfo(&info->default_actions);
+    InitVMods(&info->mods, mods, include_depth > 0);
     info->explicit_group = XKB_LAYOUT_INVALID;
 }
 
@@ -498,8 +451,7 @@ UseNewKeyField(enum key_field field, enum key_field old, enum key_field new,
         if (report)
             *collide |= field;
 
-        if (clobber)
-            return true;
+        return clobber;
     }
 
     return false;
@@ -652,7 +604,7 @@ MergeIncludedSymbols(SymbolsInfo *into, SymbolsInfo *from,
         return;
     }
 
-    into->mods = from->mods;
+    MergeModSets(into->ctx, &into->mods, &from->mods, merge);
 
     if (into->name == NULL) {
         into->name = steal(&from->name);
@@ -680,7 +632,7 @@ MergeIncludedSymbols(SymbolsInfo *into, SymbolsInfo *from,
     else {
         KeyInfo *keyi;
         darray_foreach(keyi, from->keys) {
-            keyi->merge = (merge == MERGE_DEFAULT ? keyi->merge : merge);
+            keyi->merge = merge;
             if (!AddKeySymbols(into, keyi, false))
                 into->errorCount++;
         }
@@ -693,7 +645,7 @@ MergeIncludedSymbols(SymbolsInfo *into, SymbolsInfo *from,
     else {
         ModMapEntry *mm;
         darray_foreach(mm, from->modmaps) {
-            mm->merge = (merge == MERGE_DEFAULT ? mm->merge : merge);
+            mm->merge = merge;
             if (!AddModMapEntry(into, mm))
                 into->errorCount++;
         }
@@ -701,7 +653,7 @@ MergeIncludedSymbols(SymbolsInfo *into, SymbolsInfo *from,
 }
 
 static void
-HandleSymbolsFile(SymbolsInfo *info, XkbFile *file, enum merge_mode merge);
+HandleSymbolsFile(SymbolsInfo *info, XkbFile *file);
 
 static bool
 HandleIncludeSymbols(SymbolsInfo *info, IncludeStmt *include)
@@ -713,8 +665,8 @@ HandleIncludeSymbols(SymbolsInfo *info, IncludeStmt *include)
         return false;
     }
 
-    InitSymbolsInfo(&included, info->keymap, 0 /* unused */,
-                    info->actions, &info->mods);
+    InitSymbolsInfo(&included, info->keymap, info->include_depth + 1,
+                    &info->mods);
     included.name = steal(&include->stmt);
 
     for (IncludeStmt *stmt = include; stmt; stmt = stmt->next_incl) {
@@ -729,7 +681,7 @@ HandleIncludeSymbols(SymbolsInfo *info, IncludeStmt *include)
         }
 
         InitSymbolsInfo(&next_incl, info->keymap, info->include_depth + 1,
-                        info->actions, &included.mods);
+                        &included.mods);
         if (stmt->modifier) {
             next_incl.explicit_group = atoi(stmt->modifier) - 1;
             if (next_incl.explicit_group >= XKB_MAX_GROUPS) {
@@ -754,7 +706,7 @@ HandleIncludeSymbols(SymbolsInfo *info, IncludeStmt *include)
             next_incl.explicit_group = info->explicit_group;
         }
 
-        HandleSymbolsFile(&next_incl, file, MERGE_OVERRIDE);
+        HandleSymbolsFile(&next_incl, file);
 
         MergeIncludedSymbols(&included, &next_incl, stmt->merge);
 
@@ -818,25 +770,23 @@ static bool
 AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
                 ExprDef *value)
 {
-    xkb_layout_index_t ndx;
-    GroupInfo *groupi;
-    xkb_level_index_t nLevels;
+    xkb_layout_index_t ndx = 0;
 
     if (!GetGroupIndex(info, keyi, arrayNdx, GROUP_FIELD_SYMS, &ndx))
         return false;
 
-    groupi = &darray_item(keyi->groups, ndx);
+    GroupInfo *groupi = &darray_item(keyi->groups, ndx);
 
-    if (value->expr.op == EXPR_EMPTY_LIST) {
+    if (value->common.type == STMT_EXPR_EMPTY_LIST) {
         groupi->defined |= GROUP_FIELD_SYMS;
         return true;
     }
 
-    if (value->expr.op != EXPR_KEYSYM_LIST) {
+    if (value->common.type != STMT_EXPR_KEYSYM_LIST) {
         log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Expected a list of symbols, found %s; "
                 "Ignoring symbols for group %u of %s\n",
-                expr_op_type_to_string(value->expr.op), ndx + 1,
+                stmt_type_to_string(value->common.type), ndx + 1,
                 KeyInfoText(info, keyi));
         return false;
     }
@@ -849,9 +799,20 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
         return false;
     }
 
-    nLevels = 0;
-    for (ParseCommon *p = &value->common; p; p = p->next)
+    xkb_level_index_t nLevels = 0;
+    xkb_level_index_t nonEmptyLevels = 0;
+    /* Contrary to actions, keysyms are already parsed at this point so we drop
+     * trailing symbols by not adding them in the first place. */
+    for (ExprKeysymList *keysymList = (ExprKeysymList *) value;
+         keysymList;
+         keysymList = (ExprKeysymList *) keysymList->common.next) {
         nLevels++;
+        /* Drop trailing NoSymbol */
+        if (darray_size(keysymList->syms) > 0)
+            nonEmptyLevels = nLevels;
+    }
+    if (nonEmptyLevels < nLevels)
+        nLevels = nonEmptyLevels;
     if (darray_size(groupi->levels) < nLevels)
         darray_resize0(groupi->levels, nLevels);
 
@@ -859,12 +820,21 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
 
     xkb_level_index_t level = 0;
     for (ExprKeysymList *keysymList = (ExprKeysymList *) value;
-         keysymList;
-         keysymList = (ExprKeysymList *) keysymList->expr.common.next, level++) {
+         keysymList && level < nLevels;
+         keysymList = (ExprKeysymList *) keysymList->common.next, level++) {
         struct xkb_level *leveli = &darray_item(groupi->levels, level);
         assert(leveli->num_syms == 0);
 
-        leveli->num_syms = darray_size(keysymList->syms);
+        if (unlikely(darray_size(keysymList->syms) > MAX_KEYSYMS_PER_LEVEL)) {
+            log_err(info->ctx, XKB_LOG_MESSAGE_NO_ID,
+                    "Key %s has too many keysyms for group %u, level %u; "
+                    "expected max %u, got: %u\n",
+                    KeyInfoText(info, keyi), ndx + 1, level + 1,
+                    MAX_KEYSYMS_PER_LEVEL, darray_size(keysymList->syms));
+            return false;
+        }
+
+        leveli->num_syms = (xkb_keysym_count_t) darray_size(keysymList->syms);
         switch (leveli->num_syms) {
         case 0:
             leveli->s.sym = XKB_KEY_NoSymbol;
@@ -874,20 +844,11 @@ AddSymbolsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
             assert(leveli->s.sym != XKB_KEY_NoSymbol);
             break;
         default:
-            leveli->s.syms =
-                calloc(leveli->num_syms, sizeof(*leveli->s.syms));
-            if (!leveli->s.syms) {
-                leveli->num_syms = 0;
-                log_err(info->ctx, XKB_ERROR_ALLOCATION_ERROR,
-                        "Could not allocate keysyms\n");
-                return false;
-            }
-            memcpy(leveli->s.syms,
-                   &darray_item(keysymList->syms, 0),
-                   leveli->num_syms * sizeof(*leveli->s.syms));
+            darray_shrink(keysymList->syms);
+            darray_steal(keysymList->syms, &leveli->s.syms, NULL);
 #ifndef NDEBUG
             /* Canonical list: all NoSymbol were dropped */
-            for (unsigned int k = 0; k < leveli->num_syms; k++)
+            for (xkb_keysym_count_t k = 0; k < leveli->num_syms; k++)
                 assert(leveli->s.syms[k] != XKB_KEY_NoSymbol);
 #endif
         }
@@ -900,25 +861,23 @@ static bool
 AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
                 ExprDef *value)
 {
-    xkb_layout_index_t ndx;
-    GroupInfo *groupi;
-    xkb_level_index_t nLevels;
+    xkb_layout_index_t ndx = 0;
 
     if (!GetGroupIndex(info, keyi, arrayNdx, GROUP_FIELD_ACTS, &ndx))
         return false;
 
-    groupi = &darray_item(keyi->groups, ndx);
+    GroupInfo *groupi = &darray_item(keyi->groups, ndx);
 
-    if (value->expr.op == EXPR_EMPTY_LIST) {
+    if (value->common.type == STMT_EXPR_EMPTY_LIST) {
         groupi->defined |= GROUP_FIELD_ACTS;
         return true;
     }
 
-    if (value->expr.op != EXPR_ACTION_LIST) {
+    if (value->common.type != STMT_EXPR_ACTION_LIST) {
         log_wsgo(info->ctx, XKB_ERROR_INVALID_EXPRESSION_TYPE,
                  "Bad expression type (%d) for action list value; "
                  "Ignoring actions for group %u of %s\n",
-                 value->expr.op, ndx, KeyInfoText(info, keyi));
+                 value->common.type, ndx, KeyInfoText(info, keyi));
         return false;
     }
 
@@ -929,7 +888,10 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
         return false;
     }
 
-    nLevels = 0;
+    xkb_level_index_t nLevels = 0;
+    /* Contrary to keysyms with trailing `NoSymbol`, we cannot detect trailing
+     * `NoAction()` now, because we need to parse the actions first. Jusr count
+     * explicit actions for now. */
     for (ParseCommon *p = &value->common; p; p = p->next)
         nLevels++;
     if (darray_size(groupi->levels) < nLevels)
@@ -938,22 +900,35 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
     groupi->defined |= GROUP_FIELD_ACTS;
 
     xkb_level_index_t level = 0;
+    xkb_level_index_t nonEmptyLevels = 0;
     for (ExprActionList *actionList = (ExprActionList *) value;
          actionList;
-         actionList = (ExprActionList *) actionList->expr.common.next, level++) {
+         actionList = (ExprActionList *) actionList->common.next, level++) {
         struct xkb_level *leveli = &darray_item(groupi->levels, level);
         assert(leveli->num_actions == 0);
 
         unsigned int num_actions = 0;
-        for (ParseCommon *p = &actionList->actions->common; p; p = p->next)
+        for (ExprDef *act = actionList->actions;
+             act; act = (ExprDef *) act->common.next)
             num_actions++;
 
+        if (unlikely(num_actions > MAX_ACTIONS_PER_LEVEL)) {
+            log_err(info->ctx, XKB_LOG_MESSAGE_NO_ID,
+                    "Key %s has too many actions for group %u, level %u; "
+                    "expected max %u, got: %u\n",
+                    KeyInfoText(info, keyi), ndx + 1, level + 1,
+                    MAX_ACTIONS_PER_LEVEL, num_actions);
+            return false;
+        }
+
         /* Parse actions and add only defined actions */
         darray(union xkb_action) actions = darray_new();
         unsigned int act_index = 0;
-        for (ExprDef *act = actionList->actions; act; act = (ExprDef *) act->common.next, act_index++) {
+        for (ExprDef *act = actionList->actions;
+             act; act = (ExprDef *) act->common.next, act_index++) {
             union xkb_action toAct = { 0 };
-            if (!HandleActionDef(info->ctx, info->actions, &info->mods, act, &toAct))
+            if (!HandleActionDef(info->ctx, &info->default_actions, &info->mods,
+                                 act, &toAct))
                 log_err(info->ctx, XKB_ERROR_INVALID_VALUE,
                         "Illegal action definition for %s; "
                         "Action for group %u/level %u ignored\n",
@@ -972,12 +947,12 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
         if (darray_empty(actions)) {
             leveli->num_actions = 0;
         } else if (likely(darray_size(actions) > 1)) {
-            leveli->num_actions = darray_size(actions);
+            leveli->num_actions = (xkb_action_count_t) darray_size(actions);
             darray_shrink(actions);
             darray_steal(actions, &leveli->a.actions, NULL);
 #ifndef NDEBUG
             /* Canonical list: all NoAction() were dropped */
-            for (unsigned int k = 0; k < leveli->num_actions; k++)
+            for (xkb_action_count_t k = 0; k < leveli->num_actions; k++)
                 assert(leveli->a.actions[k].type != ACTION_TYPE_NONE);
 #endif
         } else {
@@ -989,7 +964,22 @@ AddActionsToKey(SymbolsInfo *info, KeyInfo *keyi, ExprDef *arrayNdx,
             darray_free(actions);
         }
 next_level:
-        ;
+        /*
+         * Check trailing `NoAction()`, but count as empty level only if there
+         * is no corresponding keysym.
+         */
+        if (leveli->num_actions > 0 || leveli->num_syms > 0)
+            nonEmptyLevels = level + 1;
+    }
+
+    if (nonEmptyLevels < nLevels) {
+        /* Drop trailing `NoAction()`.
+         * No need to clear dropped levels: there are no keysyms nor actions */
+        darray_size(groupi->levels) = nonEmptyLevels;
+        if (nonEmptyLevels > 0)
+            darray_shrink(groupi->levels);
+        else
+            darray_free(groupi->levels);
     }
 
     return true;
@@ -1011,8 +1001,8 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
                 ExprDef *arrayNdx, ExprDef *value)
 {
     if (istreq(field, "type")) {
-        xkb_layout_index_t ndx;
-        xkb_atom_t val;
+        xkb_layout_index_t ndx = 0;
+        xkb_atom_t val = XKB_ATOM_NONE;
 
         if (!ExprResolveString(info->ctx, value, &val)) {
             log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
@@ -1049,14 +1039,14 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
     else if (istreq(field, "vmods") ||
              istreq(field, "virtualmods") ||
              istreq(field, "virtualmodifiers")) {
-        xkb_mod_mask_t mask;
+        xkb_mod_mask_t mask = 0;
 
         if (!ExprResolveModMask(info->ctx, value, MOD_VIRT, &info->mods,
                                 &mask)) {
             log_err(info->ctx, XKB_ERROR_UNSUPPORTED_MODIFIER_MASK,
                     "Expected a virtual modifier mask, found %s; "
                     "Ignoring virtual modifiers definition for key %s\n",
-                    expr_op_type_to_string(value->expr.op),
+                    stmt_type_to_string(value->common.type),
                     KeyInfoText(info, keyi));
             return false;
         }
@@ -1090,7 +1080,7 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
     else if (istreq(field, "repeating") ||
              istreq(field, "repeats") ||
              istreq(field, "repeat")) {
-        unsigned int val;
+        uint32_t val = 0;
 
         if (!ExprResolveEnum(info->ctx, value, &val, repeatEntries)) {
             log_err(info->ctx, XKB_ERROR_INVALID_VALUE,
@@ -1100,12 +1090,12 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
             return false;
         }
 
-        keyi->repeat = val;
+        keyi->repeat = (enum key_repeat) val;
         keyi->defined |= KEY_FIELD_REPEAT;
     }
     else if (istreq(field, "groupswrap") ||
              istreq(field, "wrapgroups")) {
-        bool set;
+        bool set = false;
 
         if (!ExprResolveBoolean(info->ctx, value, &set)) {
             log_err(info->ctx, XKB_ERROR_INVALID_VALUE,
@@ -1120,7 +1110,7 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
     }
     else if (istreq(field, "groupsclamp") ||
              istreq(field, "clampgroups")) {
-        bool set;
+        bool set = false;
 
         if (!ExprResolveBoolean(info->ctx, value, &set)) {
             log_err(info->ctx, XKB_ERROR_INVALID_VALUE,
@@ -1135,7 +1125,7 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
     }
     else if (istreq(field, "groupsredirect") ||
              istreq(field, "redirectgroups")) {
-        xkb_layout_index_t grp;
+        xkb_layout_index_t grp = 0;
 
         if (!ExprResolveGroup(info->ctx, value, &grp)) {
             log_err(info->ctx, XKB_ERROR_UNSUPPORTED_GROUP_INDEX,
@@ -1151,7 +1141,7 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
     }
     else {
         log_err(info->ctx, XKB_ERROR_UNKNOWN_FIELD,
-                "Unknown field %s in a symbol interpretation; "
+                "Unknown field \"%s\" in a symbol interpretation; "
                 "Definition ignored\n",
                 field);
         return false;
@@ -1161,11 +1151,9 @@ SetSymbolsField(SymbolsInfo *info, KeyInfo *keyi, const char *field,
 }
 
 static bool
-SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value)
+SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value,
+             enum merge_mode merge)
 {
-    xkb_layout_index_t group, group_to_use;
-    xkb_atom_t name;
-
     if (!arrayNdx) {
         log_vrb(info->ctx, 1, XKB_WARNING_MISSING_SYMBOLS_GROUP_NAME_INDEX,
                 "You must specify an index when specifying a group name; "
@@ -1173,6 +1161,7 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value)
         return false;
     }
 
+    xkb_layout_index_t group = 0;
     if (!ExprResolveGroup(info->ctx, arrayNdx, &group)) {
         log_err(info->ctx, XKB_ERROR_UNSUPPORTED_GROUP_INDEX,
                 "Illegal index in group name definition; "
@@ -1180,6 +1169,7 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value)
         return false;
     }
 
+    xkb_atom_t name = XKB_ATOM_NONE;
     if (!ExprResolveString(info->ctx, value, &name)) {
         log_err(info->ctx, XKB_ERROR_WRONG_FIELD_TYPE,
                 "Group name must be a string; "
@@ -1187,6 +1177,7 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value)
         return false;
     }
 
+    xkb_layout_index_t group_to_use;
     if (info->explicit_group == XKB_LAYOUT_INVALID) {
         group_to_use = group - 1;
     }
@@ -1203,8 +1194,25 @@ SetGroupName(SymbolsInfo *info, ExprDef *arrayNdx, ExprDef *value)
         return false;
     }
 
-    if (group_to_use >= darray_size(info->group_names))
+    if (group_to_use >= darray_size(info->group_names)) {
+        /* Avoid clang-tidy false positive */
+        assert(darray_size(info->group_names) < group_to_use + 1);
         darray_resize0(info->group_names, group_to_use + 1);
+    } else {
+        const xkb_atom_t old_name = darray_item(info->group_names, group_to_use);
+        if (old_name != XKB_ATOM_NONE && old_name != name) {
+            const bool replace = (merge != MERGE_AUGMENT);
+            const xkb_atom_t use    = (replace ? name : old_name);
+            const xkb_atom_t ignore = (replace ? old_name : name);
+            log_warn(info->ctx, XKB_LOG_MESSAGE_NO_ID,
+                     "Multiple definitions of group %"PRIu32" name "
+                     "in map '%s'; Using '%s', ignoring '%s'\n",
+                     group_to_use, info->name,
+                     xkb_atom_text(info->ctx, use),
+                     xkb_atom_text(info->ctx, ignore));
+            name = use;
+        }
+    }
     darray_item(info->group_names, group_to_use) = name;
 
     return true;
@@ -1221,12 +1229,18 @@ HandleGlobalVar(SymbolsInfo *info, VarDef *stmt)
         return false;
 
     if (elem && istreq(elem, "key")) {
-        ret = SetSymbolsField(info, &info->default_key, field, arrayNdx,
-                              stmt->value);
+        KeyInfo temp = {0};
+        InitKeyInfo(info->ctx, &temp);
+        /* Do not replace the whole key, only the current field */
+        temp.merge = (temp.merge == MERGE_REPLACE)
+            ? MERGE_OVERRIDE
+            : stmt->merge;
+        ret = SetSymbolsField(info, &temp, field, arrayNdx, stmt->value);
+        MergeKeys(info, &info->default_key, &temp, true);
     }
     else if (!elem && (istreq(field, "name") ||
                        istreq(field, "groupname"))) {
-        ret = SetGroupName(info, arrayNdx, stmt->value);
+        ret = SetGroupName(info, arrayNdx, stmt->value, stmt->merge);
     }
     else if (!elem && (istreq(field, "groupswrap") ||
                        istreq(field, "wrapgroups"))) {
@@ -1253,8 +1267,9 @@ HandleGlobalVar(SymbolsInfo *info, VarDef *stmt)
         ret = true;
     }
     else {
-        ret = SetActionField(info->ctx, info->actions, &info->mods,
-                             elem, field, arrayNdx, stmt->value);
+        ret = SetDefaultActionField(info->ctx, &info->default_actions,
+                                    &info->mods, elem, field, arrayNdx,
+                                    stmt->value, stmt->merge);
     }
 
     return ret;
@@ -1274,7 +1289,7 @@ HandleSymbolsBody(SymbolsInfo *info, VarDef *def, KeyInfo *keyi)
         bool ok = true;
 
         if (!def->name) {
-            if (unlikely(!def->value) || def->value->expr.op != EXPR_ACTION_LIST)
+            if (unlikely(!def->value) || def->value->common.type != STMT_EXPR_ACTION_LIST)
                 field = "symbols"; /* Default to symbols field */
             else
                 field = "actions";
@@ -1358,22 +1373,15 @@ HandleSymbolsDef(SymbolsInfo *info, SymbolsDef *stmt)
     keyi.merge = stmt->merge;
     keyi.name = stmt->keyName;
 
-    if (!HandleSymbolsBody(info, stmt->symbols, &keyi)) {
-        info->errorCount++;
-        return false;
+    if (HandleSymbolsBody(info, stmt->symbols, &keyi) &&
+        SetExplicitGroup(info, &keyi) &&
+        AddKeySymbols(info, &keyi, true)) {
+        return true;
     }
 
-    if (!SetExplicitGroup(info, &keyi)) {
-        info->errorCount++;
-        return false;
-    }
-
-    if (!AddKeySymbols(info, &keyi, true)) {
-        info->errorCount++;
-        return false;
-    }
-
-    return true;
+    ClearKeyInfo(&keyi);
+    info->errorCount++;
+    return false;
 }
 
 static bool
@@ -1386,10 +1394,10 @@ HandleModMapDef(SymbolsInfo *info, ModMapDef *def)
     const char *modifier_name = xkb_atom_text(ctx, def->modifier);
 
     if (istreq(modifier_name, "none")) {
-        // Handle special "None" entry
+        /* Handle special "None" entry */
         ndx = XKB_MOD_NONE;
     } else {
-        // Handle normal entry
+        /* Handle normal entry */
         ndx = XkbModNameToIndex(&info->mods, def->modifier, MOD_REAL);
         if (ndx == XKB_MOD_INVALID) {
             log_err(info->ctx, XKB_ERROR_INVALID_REAL_MODIFIER,
@@ -1405,16 +1413,18 @@ HandleModMapDef(SymbolsInfo *info, ModMapDef *def)
     tmp.merge = def->merge;
 
     for (ExprDef *key = def->keys; key; key = (ExprDef *) key->common.next) {
-        xkb_keysym_t sym;
-
-        if (key->expr.op == EXPR_VALUE &&
-            key->expr.value_type == EXPR_TYPE_KEYNAME) {
+        if (key->common.type == STMT_EXPR_KEYNAME_LITERAL) {
             tmp.haveSymbol = false;
             tmp.u.keyName = key->key_name.key_name;
         }
-        else if (ExprResolveKeySym(ctx, key, &sym)) {
-            tmp.haveSymbol = true;
-            tmp.u.keySym = sym;
+        else if (key->common.type == STMT_EXPR_KEYSYM_LITERAL) {
+            if (key->keysym.keysym == XKB_KEY_NoSymbol) {
+                /* Invalid keysym: ignore. Error message already printed */
+                continue;
+            } else {
+                tmp.haveSymbol = true;
+                tmp.u.keySym = key->keysym.keysym;
+            }
         }
         else {
             log_err(info->ctx, XKB_ERROR_INVALID_MODMAP_ENTRY,
@@ -1430,7 +1440,7 @@ HandleModMapDef(SymbolsInfo *info, ModMapDef *def)
 }
 
 static void
-HandleSymbolsFile(SymbolsInfo *info, XkbFile *file, enum merge_mode merge)
+HandleSymbolsFile(SymbolsInfo *info, XkbFile *file)
 {
     bool ok;
 
@@ -1449,7 +1459,7 @@ HandleSymbolsFile(SymbolsInfo *info, XkbFile *file, enum merge_mode merge)
             ok = HandleGlobalVar(info, (VarDef *) stmt);
             break;
         case STMT_VMOD:
-            ok = HandleVModDef(info->ctx, &info->mods, (VModDef *) stmt, merge);
+            ok = HandleVModDef(info->ctx, &info->mods, (VModDef *) stmt);
             break;
         case STMT_MODMAP:
             ok = HandleModMapDef(info, (ModMapDef *) stmt);
@@ -1466,9 +1476,9 @@ HandleSymbolsFile(SymbolsInfo *info, XkbFile *file, enum merge_mode merge)
             info->errorCount++;
 
         if (info->errorCount > 10) {
-            log_err(info->ctx, XKB_ERROR_INVALID_SYNTAX,
+            log_err(info->ctx, XKB_ERROR_INVALID_XKB_SYNTAX,
                     "Abandoning symbols file \"%s\"\n",
-                    file->name);
+                    safe_map_name(file));
             break;
         }
     }
@@ -1598,7 +1608,7 @@ FindTypeForGroup(struct xkb_keymap *keymap, KeyInfo *keyi,
 
     if (type_name == XKB_ATOM_NONE) {
         if (keyi->default_type != XKB_ATOM_NONE) {
-            type_name  = keyi->default_type;
+            type_name = keyi->default_type;
         }
         else {
             type_name = FindAutomaticType(keymap->ctx, groupi);
@@ -1609,8 +1619,8 @@ FindTypeForGroup(struct xkb_keymap *keymap, KeyInfo *keyi,
 
     if (type_name == XKB_ATOM_NONE) {
         log_warn(keymap->ctx, XKB_WARNING_CANNOT_INFER_KEY_TYPE,
-                 "Couldn't find an automatic type for key '%s' group %d with %lu levels; "
-                 "Using the default type\n",
+                 "Couldn't find an automatic type for key '%s' "
+                 "group %d with %lu levels; Using the default type\n",
                  KeyNameText(keymap->ctx, keyi->name), group + 1,
                  (unsigned long) darray_size(groupi->levels));
         goto use_default;
@@ -1622,8 +1632,8 @@ FindTypeForGroup(struct xkb_keymap *keymap, KeyInfo *keyi,
 
     if (i >= keymap->num_types) {
         log_warn(keymap->ctx, XKB_WARNING_UNDEFINED_KEY_TYPE,
-                 "The type \"%s\" for key '%s' group %d was not previously defined; "
-                 "Using the default type\n",
+                 "The type \"%s\" for key '%s' group %d "
+                 "was not previously defined; Using the default type\n",
                  xkb_atom_text(keymap->ctx, type_name),
                  KeyNameText(keymap->ctx, keyi->name), group + 1);
         goto use_default;
@@ -1662,12 +1672,24 @@ CopySymbolsDefToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info,
 
     /* Find the range of groups we need. */
     key->num_groups = 0;
-    darray_enumerate(i, groupi, keyi->groups)
-        if (groupi->defined)
+    darray_enumerate(i, groupi, keyi->groups) {
+        /* Skip groups that have no levels and no explicit type.
+         * Such group would be filled with `NoSymbol` anyway. */
+        const bool has_explicit_type = (keyi->defined & KEY_FIELD_DEFAULT_TYPE)
+                                    || (groupi->defined & GROUP_FIELD_TYPE);
+        if (darray_size(groupi->levels) > 0 || has_explicit_type)
             key->num_groups = i + 1;
+        if (has_explicit_type)
+            key->explicit |= EXPLICIT_TYPES;
+    }
 
-    if (key->num_groups <= 0)
-        return false; /* WSGO */
+    if (key->num_groups <= 0) {
+        /* A key with no group may still have other fields defined */
+        if (keyi->defined)
+            goto key_fields;
+        else
+            return false;
+    }
 
     darray_resize(keyi->groups, key->num_groups);
 
@@ -1698,14 +1720,14 @@ CopySymbolsDefToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info,
             struct xkb_level *leveli;
 
             log_vrb(info->ctx, 1, XKB_WARNING_EXTRA_SYMBOLS_IGNORED,
-                    "Type \"%s\" has %d levels, but %s has %d levels; "
+                    "Type \"%s\" has %d levels, but %s has %u levels; "
                     "Ignoring extra symbols\n",
                     xkb_atom_text(keymap->ctx, type->name), type->num_levels,
                     KeyInfoText(info, keyi),
-                    (int) darray_size(groupi->levels));
+                    darray_size(groupi->levels));
 
             darray_foreach_from(leveli, groupi->levels, type->num_levels)
-                ClearLevelInfo(leveli);
+                clear_level(leveli);
         }
         darray_resize0(groupi->levels, type->num_levels);
 
@@ -1715,16 +1737,68 @@ CopySymbolsDefToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info,
 
     /* Copy levels. */
     darray_enumerate(i, groupi, keyi->groups) {
+        /*
+         * Compute the capitalization transformation of the keysyms.
+         * This is necessary because `xkb_state_key_get_syms()` returns an
+         * immutable array and does not use a buffer, so we must store them.
+         * We apply only simple capitalization rules, so the keysyms count is
+         * unchanged.
+         */
+        struct xkb_level *leveli;
+        darray_foreach(leveli, groupi->levels) {
+            switch (leveli->num_syms) {
+            case 0:
+                leveli->upper = XKB_KEY_NoSymbol;
+                break;
+            case 1:
+                leveli->upper = xkb_keysym_to_upper(leveli->s.sym);
+                break;
+            default:
+                /* Multiple keysyms: check if there is any cased keysym */
+                leveli->has_upper = false;
+                for (xkb_keysym_count_t k = 0; k < leveli->num_syms; k++) {
+                    const xkb_keysym_t upper =
+                        xkb_keysym_to_upper(leveli->s.syms[k]);
+                    if (upper != leveli->s.syms[k]) {
+                        leveli->has_upper = true;
+                        break;
+                    }
+                }
+                if (leveli->has_upper) {
+                    /*
+                     * Some cased keysyms: store the transformation result in
+                     * the same array, right after the original keysyms.
+                     */
+                    leveli->s.syms = realloc(leveli->s.syms,
+                                             (size_t) 2 * leveli->num_syms *
+                                             sizeof(*leveli->s.syms));
+                    if (!leveli->s.syms)
+                        return false; /* FIXME: better handling? */
+                    for (xkb_keysym_count_t k = 0; k < leveli->num_syms; k++) {
+                        leveli->s.syms[leveli->num_syms + k] =
+                            xkb_keysym_to_upper(leveli->s.syms[k]);
+                    }
+                }
+            }
+        }
+
+        /* Copy the level */
         darray_steal(groupi->levels, &key->groups[i].levels, NULL);
-        if (groupi->defined & GROUP_FIELD_ACTS) {
+        if (key->groups[i].type->num_levels > 1 ||
+            key->groups[i].levels[0].num_syms > 0)
+            key->explicit |= EXPLICIT_SYMBOLS;
+        if (groupi->defined & GROUP_FIELD_ACTS) { // FIXME
             key->groups[i].explicit_actions = true;
             key->explicit |= EXPLICIT_INTERP;
         }
+        if (key->groups[i].explicit_type)
+            key->explicit |= EXPLICIT_TYPES;
     }
 
     key->out_of_range_group_number = keyi->out_of_range_group_number;
     key->out_of_range_group_action = keyi->out_of_range_group_action;
 
+key_fields:
     if (keyi->defined & KEY_FIELD_VMODMAP) {
         key->vmodmap = keyi->vmodmap;
         key->explicit |= EXPLICIT_VMODMAP;
@@ -1767,9 +1841,9 @@ CopyModMapDefToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info,
         }
     }
 
-    // Handle modMap None
+    /* Skip modMap None */
     if (entry->modifier != XKB_MOD_NONE) {
-        // Convert modifier index to modifier mask
+        /* Convert modifier index to modifier mask */
         key->modmap |= (UINT32_C(1) << entry->modifier);
     }
 
@@ -1817,20 +1891,14 @@ CopySymbolsToKeymap(struct xkb_keymap *keymap, SymbolsInfo *info)
 }
 
 bool
-CompileSymbols(XkbFile *file, struct xkb_keymap *keymap,
-               enum merge_mode merge)
+CompileSymbols(XkbFile *file, struct xkb_keymap *keymap)
 {
     SymbolsInfo info;
-    ActionsInfo *actions;
 
-    actions = NewActionsInfo();
-    if (!actions)
-        return false;
+    InitSymbolsInfo(&info, keymap, 0, &keymap->mods);
 
-    InitSymbolsInfo(&info, keymap, 0, actions, &keymap->mods);
-    info.default_key.merge = merge;
-
-    HandleSymbolsFile(&info, file, merge);
+    if (file !=NULL)
+        HandleSymbolsFile(&info, file);
 
     if (info.errorCount != 0)
         goto err_info;
@@ -1839,11 +1907,9 @@ CompileSymbols(XkbFile *file, struct xkb_keymap *keymap,
         goto err_info;
 
     ClearSymbolsInfo(&info);
-    FreeActionsInfo(actions);
     return true;
 
 err_info:
-    FreeActionsInfo(actions);
     ClearSymbolsInfo(&info);
     return false;
 }
