@@ -1,3 +1,9 @@
+/*
+ * Copyright © 2024 Pierre Le Marre <dev@wismill.eu>
+ * SPDX-License-Identifier: MIT
+ */
+#pragma once
+
 #include <stddef.h>
 #include <stdint.h>
 
